import { AlgorithmToLayoutProvider, AllLayoutAlgorithms, LayoutFactory } from '@graphpolaris/shared/lib/graph-layout';
import { schemaGraphology2Reactflow, schemaExpandRelation } from '@graphpolaris/shared/lib/schema/schema-utils';
import {
  useSchemaGraph,
  useSchemaLayout,
  useSchemaSettings,
  useSearchResultSchema,
  useSessionCache,
} from '@graphpolaris/shared/lib/data-access/store';
import { SmartBezierEdge, SmartStepEdge, SmartStraightEdge } from '@tisoap/react-flow-smart-edge';

import { useEffect, useMemo, useRef, useState } from 'react';
import ReactFlow, {
  ControlButton,
  Controls,
  Node,
  Edge,
  ReactFlowProvider,
  useNodesState,
  useEdgesState,
  ReactFlowInstance,
  useReactFlow,
} from 'reactflow';
import CachedIcon from '@mui/icons-material/Cached';
import SettingsIcon from '@mui/icons-material/Settings';

import 'reactflow/dist/style.css';

import styles from './schema.module.scss';

import { ConnectionDragLine, ConnectionLine } from '@graphpolaris/shared/lib/querybuilder/pills';
import { EntityNode } from '../pills/nodes/entity/entity-node';
import { RelationNode } from '../pills/nodes/relation/relation-node';
import NodeEdge from '../pills/edges/node-edge';
import SelfEdge from '../pills/edges/self-edge';
import { useSchemaAPI } from '../../data-access';
import { SchemaDialog } from './schemaDialog';
import { toSchemaGraphology } from '../../data-access/store/schemaSlice';

interface Props {
  content?: string;
  auth?: boolean;
}

const onInit = (reactFlowInstance: ReactFlowInstance) => {
  setTimeout(() => reactFlowInstance.fitView(), 100);
};

const nodeTypes = {
  entity: EntityNode,
  relation: RelationNode,
  // nodeQualityEntityPopup: NodeQualityEntityPopupNode,
  // nodeQualityRelationPopup: NodeQualityRelationPopupNode,
  // attributeAnalyticsPopupMenu: AttributeAnalyticsPopupMenu,

  // entity: EntityRFPill,
  // relation: RelationRFPill,
  // attribute: AttributeRFPill,
};
const edgeTypes = {
  nodeEdge: NodeEdge,
  selfEdge: SelfEdge,
  // connection: ConnectionLine,
  bezier: SmartBezierEdge,
  connection: ConnectionLine,
  straight: SmartStraightEdge,
  step: SmartStepEdge,
};

export const Schema = (props: Props) => {
  const api_schema = useSchemaAPI();
  const session = useSessionCache();
  const settings = useSchemaSettings();

  const searchResults = useSearchResultSchema();

  const [toggleSchemaSettings, setToggleSchemaSettings] = useState(false);
  const [nodes, setNodes, onNodeChanged] = useNodesState([] as Node[]);
  const [edges, setEdges, onEdgeChanged] = useEdgesState([] as Edge[]);
  const [firstUserConnection, setFirstUserConnection] = useState<boolean>(true);
  const [auth, setAuth] = useState(props.auth);
  const settingsIconRef = useRef<SVGSVGElement>(null);
  const dialogRef = useRef<HTMLDivElement>(null);

  // In case the schema is updated
  const schemaGraph = useSchemaGraph();
  const schemaGraphology = useMemo(() => toSchemaGraphology(schemaGraph), [schemaGraph]);
  const layout = useRef<AlgorithmToLayoutProvider<AllLayoutAlgorithms>>();

  const [loading, setLoading] = useState(false);

  function updateLayout() {
    const layoutFactory = new LayoutFactory();
    layout.current = layoutFactory.createLayout(settings.layoutName);
  }

  useEffect(() => {
    setLoading(true);
  }, [session.currentDatabase]);

  useEffect(() => {
    updateLayout();
    sessionStorage.setItem('firstUserConnection', firstUserConnection.toString());
  }, []);

  useEffect(() => {
    setAuth(props.auth);
  }, [props.auth]);

  useEffect(() => {
    setLoading(false);

    if (schemaGraphology === undefined || schemaGraphology.order == 0) {
      setNodes([]);
      setEdges([]);
      return;
    }

    updateLayout();
    const expandedSchema = schemaExpandRelation(schemaGraphology);
    layout.current?.layout(expandedSchema);

    const schemaFlow = schemaGraphology2Reactflow(expandedSchema, settings.connectionType);

    // schemaFlow.nodes.forEach((n) => {
    //   n.data.toggleNodeQualityPopup = toggleNodeQualityPopup;
    //   n.data.toggleAttributeAnalyticsPopupMenu =
    //     toggleAttributeAnalyticsPopupMenu;
    // });
    // console.log(edges);

    // console.log(
    //   'schema Layout',
    //   schemaLayout,
    //   'order',
    //   expandedSchema,
    //   schemaFlow
    // );
    setNodes(schemaFlow.nodes);
    setEdges(schemaFlow.edges);
    // console.log(
    //   'update schema useEffect',
    //   schemaGraphology,
    //   schemaGraphology.order,
    //   schemaFlow
    // );
  }, [schemaGraph, settings]);

  useEffect(() => {
    setNodes((nds) =>
      nds.map((node) => ({
        ...node,
        selected: searchResults.includes(node.id) || searchResults.includes(node.data.label),
      }))
    );
  }, [searchResults]);

  useEffect(() => {
    if (dialogRef.current && settingsIconRef.current) {
      dialogRef.current.style.top = `${settingsIconRef.current.getBoundingClientRect().top}px`;
      dialogRef.current.style.left = `${settingsIconRef.current.getBoundingClientRect().left + 30}px`;
    }
  }, [settingsIconRef, dialogRef, toggleSchemaSettings]);

  return (
    <div className="schema-panel w-full h-full">
      <SchemaDialog open={toggleSchemaSettings} onClose={() => setToggleSchemaSettings(false)} ref={dialogRef} />
      <div className="flex flex-col h-[1rem]">
        <h1>Schema</h1>
        {loading ? (
          <p className="text-sm">
            <svg
              aria-hidden="true"
              className="inline w-6 h-6 mr-2 text-gray-200 animate-spin fill-entity-600"
              viewBox="0 0 100 101"
              fill="none"
              xmlns="http://www.w3.org/2000/svg"
            >
              <path
                d="M100 50.5908C100 78.2051 77.6142 100.591 50 100.591C22.3858 100.591 0 78.2051 0 50.5908C0 22.9766 22.3858 0.59082 50 0.59082C77.6142 0.59082 100 22.9766 100 50.5908ZM9.08144 50.5908C9.08144 73.1895 27.4013 91.5094 50 91.5094C72.5987 91.5094 90.9186 73.1895 90.9186 50.5908C90.9186 27.9921 72.5987 9.67226 50 9.67226C27.4013 9.67226 9.08144 27.9921 9.08144 50.5908Z"
                fill="currentColor"
              />
              <path
                d="M93.9676 39.0409C96.393 38.4038 97.8624 35.9116 97.0079 33.5539C95.2932 28.8227 92.871 24.3692 89.8167 20.348C85.8452 15.1192 80.8826 10.7238 75.2124 7.41289C69.5422 4.10194 63.2754 1.94025 56.7698 1.05124C51.7666 0.367541 46.6976 0.446843 41.7345 1.27873C39.2613 1.69328 37.813 4.19778 38.4501 6.62326C39.0873 9.04874 41.5694 10.4717 44.0505 10.1071C47.8511 9.54855 51.7191 9.52689 55.5402 10.0491C60.8642 10.7766 65.9928 12.5457 70.6331 15.2552C75.2735 17.9648 79.3347 21.5619 82.5849 25.841C84.9175 28.9121 86.7997 32.2913 88.1811 35.8758C89.083 38.2158 91.5421 39.6781 93.9676 39.0409Z"
                fill="currentFill"
              />
            </svg>
            Connecting...
          </p>
        ) : (
          <p className="text-sm">{`Connected to: ${session.currentDatabase}`}</p>
        )}

        {nodes.length === 0 && !loading && <p className="text-sm">No Elements</p>}
      </div>
      <ReactFlowProvider>
        <div className="h-[calc(100%-.8rem)] w-full">
          <ReactFlow
            snapGrid={[10, 10]}
            snapToGrid
            onlyRenderVisibleElements={false}
            nodesDraggable={true}
            nodeTypes={nodeTypes}
            edgeTypes={edgeTypes}
            connectionLineComponent={ConnectionDragLine}
            onNodesChange={onNodeChanged}
            onEdgesChange={onEdgeChanged}
            nodes={nodes}
            edges={edges}
            onInit={onInit}
            proOptions={{ hideAttribution: true }}
          >
            <Controls showInteractive={false} showZoom={false} showFitView={true} className={`${styles.controls} schema-settings`}>
              {/* <ControlButton
              className={styles.exportButton}
              title={'Export graph schema'}
              onClick={(event) => {
                event.stopPropagation();
                // this.setState({
                //   ...this.state,
                //   exportMenuAnchor: event.currentTarget,
                // });
              }}
            >
              <img src={exportIcon} width={21}></img>
            </ControlButton> */}
              <ControlButton
                className={styles.exportButton}
                title={'Refresh graph schema'}
                onClick={(event) => {
                  event.stopPropagation();
                  api_schema.RequestSchema(session.currentDatabase);
                }}
              >
                <CachedIcon />
              </ControlButton>
              <ControlButton
                className={styles.exportButton}
                title={'Open Settings'}
                onClick={(event) => {
                  event.stopPropagation();
                  setToggleSchemaSettings(!toggleSchemaSettings);
                }}
              >
                <SettingsIcon ref={settingsIconRef} />
              </ControlButton>
            </Controls>
          </ReactFlow>
        </div>
      </ReactFlowProvider>
    </div>
  );
};

export default Schema;
